@@ -1,4 +1,7 @@
-/* Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
+/* SPDX-License-Identifier: GPL-2.0
+ *
+ * Copyright (C) 2015-2018 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
+ */
 
 #include "device.h"
 #include "peer.h"
@@ -20,7 +23,6 @@ static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct
 		.saddr = endpoint->src4.s_addr,
 		.daddr = endpoint->addr4.sin_addr.s_addr,
 		.fl4_dport = endpoint->addr4.sin_port,
-		.fl4_sport = htons(wg->incoming_port),
 		.flowi4_mark = wg->fwmark,
 		.flowi4_proto = IPPROTO_UDP
 	};
@@ -40,6 +42,8 @@ static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct
 		goto err;
 	}
 
+	fl.fl4_sport = inet_sk(sock)->inet_sport;
+
 	if (cache)
 		rt = dst_cache_get_ip4(cache, &fl.saddr);
 
@@ -89,7 +93,6 @@ static inline int send6(struct wireguard_device *wg, struct sk_buff *skb, struct
 		.saddr = endpoint->src6,
 		.daddr = endpoint->addr6.sin6_addr,
 		.fl6_dport = endpoint->addr6.sin6_port,
-		.fl6_sport = htons(wg->incoming_port),
 		.flowi6_mark = wg->fwmark,
 		.flowi6_oif = endpoint->addr6.sin6_scope_id,
 		.flowi6_proto = IPPROTO_UDP
@@ -111,6 +114,8 @@ static inline int send6(struct wireguard_device *wg, struct sk_buff *skb, struct
 		goto err;
 	}
 
+	fl.fl6_sport = inet_sk(sock)->inet_sport;
+
 	if (cache)
 		dst = dst_cache_get_ip6(cache, &fl.saddr);
 
@@ -173,6 +178,7 @@ int socket_send_buffer_to_peer(struct wireguard_peer *peer, void *buffer, size_t
 		return -ENOMEM;
 
 	skb_reserve(skb, SKB_HEADER_LEN);
+	skb_set_inner_network_header(skb, 0);
 	memcpy(skb_put(skb, len), buffer, len);
 	return socket_send_skb_to_peer(peer, skb, ds);
 }
@@ -193,14 +199,14 @@ int socket_send_buffer_as_reply_to_skb(struct wireguard_device *wg, struct sk_bu
 	if (unlikely(!skb))
 		return -ENOMEM;
 	skb_reserve(skb, SKB_HEADER_LEN);
+	skb_set_inner_network_header(skb, 0);
 	memcpy(skb_put(skb, len), out_buffer, len);
 
 	if (endpoint.addr.sa_family == AF_INET)
 		ret = send4(wg, skb, &endpoint, 0, NULL);
 	else if (endpoint.addr.sa_family == AF_INET6)
 		ret = send6(wg, skb, &endpoint, 0, NULL);
-	else
-		ret = -EAFNOSUPPORT;
+	/* No other possibilities if the endpoint is valid, which it is, as we checked above. */
 
 	return ret;
 }
@@ -241,7 +247,8 @@ void socket_set_peer_endpoint(struct wireguard_peer *peer, const struct endpoint
 	/* First we check unlocked, in order to optimize, since it's pretty rare
 	 * that an endpoint will change. If we happen to be mid-write, and two
 	 * CPUs wind up writing the same thing or something slightly different,
-	 * it doesn't really matter much either. */
+	 * it doesn't really matter much either.
+	 */
 	if (endpoint_eq(endpoint, &peer->endpoint))
 		return;
 	write_lock_bh(&peer->endpoint_lock);
@@ -262,6 +269,7 @@ out:
 void socket_set_peer_endpoint_from_skb(struct wireguard_peer *peer, const struct sk_buff *skb)
 {
 	struct endpoint endpoint;
+
 	if (!socket_endpoint_from_skb(&endpoint, skb))
 		socket_set_peer_endpoint(peer, &endpoint);
 }
@@ -306,87 +314,75 @@ static inline void set_sock_opts(struct socket *sock)
 	sk_set_memalloc(sock->sk);
 }
 
-int socket_init(struct wireguard_device *wg)
+int socket_init(struct wireguard_device *wg, u16 port)
 {
-	int ret = 0;
+	int ret;
 	struct udp_tunnel_sock_cfg cfg = {
 		.sk_user_data = wg,
 		.encap_type = 1,
 		.encap_rcv = receive
 	};
-	struct socket *new4 = NULL;
+	struct socket *new4 = NULL, *new6 = NULL;
 	struct udp_port_cfg port4 = {
 		.family = AF_INET,
 		.local_ip.s_addr = htonl(INADDR_ANY),
-		.local_udp_port = htons(wg->incoming_port),
+		.local_udp_port = htons(port),
 		.use_udp_checksums = true
 	};
 #if IS_ENABLED(CONFIG_IPV6)
 	int retries = 0;
-	struct socket *new6 = NULL;
 	struct udp_port_cfg port6 = {
 		.family = AF_INET6,
 		.local_ip6 = IN6ADDR_ANY_INIT,
-		.local_udp_port = htons(wg->incoming_port),
 		.use_udp6_tx_checksums = true,
 		.use_udp6_rx_checksums = true,
 		.ipv6_v6only = true
 	};
 #endif
 
-	mutex_lock(&wg->socket_update_lock);
 #if IS_ENABLED(CONFIG_IPV6)
 retry:
 #endif
-	if (rcu_dereference_protected(wg->sock4, lockdep_is_held(&wg->socket_update_lock)) || rcu_dereference_protected(wg->sock6, lockdep_is_held(&wg->socket_update_lock))) {
-		ret = -EADDRINUSE;
-		goto out;
-	}
 
 	ret = udp_sock_create(wg->creating_net, &port4, &new4);
 	if (ret < 0) {
 		pr_err("%s: Could not create IPv4 socket\n", wg->dev->name);
-		goto out;
+		return ret;
 	}
-	wg->incoming_port = ntohs(inet_sk(new4->sk)->inet_sport);
 	set_sock_opts(new4);
 	setup_udp_tunnel_sock(wg->creating_net, new4, &cfg);
-	rcu_assign_pointer(wg->sock4, new4->sk);
 
 #if IS_ENABLED(CONFIG_IPV6)
-	if (!ipv6_mod_enabled())
-		goto out;
-	port6.local_udp_port = htons(wg->incoming_port);
-	ret = udp_sock_create(wg->creating_net, &port6, &new6);
-	if (ret < 0) {
-		udp_tunnel_sock_release(new4);
-		rcu_assign_pointer(wg->sock4, NULL);
-		if (ret == -EADDRINUSE && !port4.local_udp_port && retries++ < 100)
-			goto retry;
-		if (!port4.local_udp_port)
-			wg->incoming_port = 0;
-		pr_err("%s: Could not create IPv6 socket\n", wg->dev->name);
-		goto out;
+	if (ipv6_mod_enabled()) {
+		port6.local_udp_port = inet_sk(new4->sk)->inet_sport;
+		ret = udp_sock_create(wg->creating_net, &port6, &new6);
+		if (ret < 0) {
+			udp_tunnel_sock_release(new4);
+			if (ret == -EADDRINUSE && !port && retries++ < 100)
+				goto retry;
+			pr_err("%s: Could not create IPv6 socket\n", wg->dev->name);
+			return ret;
+		}
+		set_sock_opts(new6);
+		setup_udp_tunnel_sock(wg->creating_net, new6, &cfg);
 	}
-	set_sock_opts(new6);
-	setup_udp_tunnel_sock(wg->creating_net, new6, &cfg);
-	rcu_assign_pointer(wg->sock6, new6->sk);
 #endif
 
-out:
-	mutex_unlock(&wg->socket_update_lock);
-	return ret;
+	socket_reinit(wg, new4 ? new4->sk : NULL, new6 ? new6->sk : NULL);
+	return 0;
 }
 
-void socket_uninit(struct wireguard_device *wg)
+void socket_reinit(struct wireguard_device *wg, struct sock *new4, struct sock *new6)
 {
 	struct sock *old4, *old6;
 
 	mutex_lock(&wg->socket_update_lock);
 	old4 = rcu_dereference_protected(wg->sock4, lockdep_is_held(&wg->socket_update_lock));
 	old6 = rcu_dereference_protected(wg->sock6, lockdep_is_held(&wg->socket_update_lock));
-	rcu_assign_pointer(wg->sock4, NULL);
-	rcu_assign_pointer(wg->sock6, NULL);
+	rcu_assign_pointer(wg->sock4, new4);
+	rcu_assign_pointer(wg->sock6, new6);
+	if (new4)
+		wg->incoming_port = ntohs(inet_sk(new4)->inet_sport);
 	mutex_unlock(&wg->socket_update_lock);
 	synchronize_rcu_bh();
 	synchronize_net();
